﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;


namespace Constellations
{
    /// <summary>
    /// Hauptfenster zum Constellation Finder.
    /// </summary>
    public partial class Hauptfenster : Form
    {
        #region Membervariablen

        /// <summary>
        /// Merker, ob die Sternbildgrenzen geladen wurden.
        /// </summary>
        bool _constellationsLoaded;
        /// <summary>
        /// Die Sternbildgrenzen.
        /// </summary>
        Hashtable _boundaries;
        /// <summary>
        /// Der Referenzpunkt für die Schnitt-Gerade.
        /// </summary>
        Coordinate _referenzpunkt;

        #endregion

        #region Konstruktion

        /// <summary>
        /// Konstruktion.
        /// </summary>
        public Hauptfenster()
        {
            InitializeComponent();

            _constellationsLoaded = false;
            _boundaries = null;
            _referenzpunkt = new Coordinate(180.0, 1000.0);
        }

        #endregion

        #region Buttonklick's

        /// <summary>
        /// Sucht anhand der eingegebenen Koordinate das Sternbid,
        /// in dem sich die Koordinate befindet.
        /// </summary>
        /// <param name="sender">Event-Quelle.</param>
        /// <param name="e">Zusätzliche Info's</param>
        private void btnSearchConstellation_Click(object sender, EventArgs e)
        {
            double ra = Double.NaN;
            double dec = Double.NaN;
            double hRa = Double.NaN;
            double mRa = Double.NaN;
            double sRa = Double.NaN;
            double gDec = Double.NaN;
            double mDec = Double.NaN;
            double sDec = Double.NaN;
            bool ok = true;
            string s = String.Empty;
            string constellation = string.Empty;
            string msg = String.Empty;

            // Auslesen des Textfeldes RA
            s = txtRA.Text.Replace(".", ",");
            ok = ok && Double.TryParse(s.Substring(0, 2), out hRa);
            ok = ok && Double.TryParse(s.Substring(4, 2), out mRa);
            ok = ok && Double.TryParse(s.Substring(8, 4), out sRa);

            // Auslesen des Textfeldes DEC
            s = txtDEC.Text.Replace(".", ",");
            ok = ok && Double.TryParse(s.Substring(0, 3), out gDec);
            ok = ok && Double.TryParse(s.Substring(5, 2), out mDec);
            ok = ok && Double.TryParse(s.Substring(8, 2), out sDec);

            if (ok)
            {
                // R Aund DEC umrechnen
                ra = ((hRa + (mRa / 60.0) + (sRa / 3600.0)) * 15.0);
                dec = Math.Sign(gDec) * (Math.Abs(gDec) + (mDec / 60.0) + (sDec / 3600.0));

                _referenzpunkt = new Coordinate(
                    ((ra > 180.0) ? -1.0 : +361.0),
                    ((dec > 0) ? -91.0 : +91.0));

                // Sternbild suchen
                ok = FindConstellation(ra, dec, ref constellation);

                // Meldung
                if (ok)
                {
                    msg = String.Format("Coordinate RA=[{0}] DEC=[{1}] is in Constellation '{2}'.",
                        txtRA.Text, txtDEC.Text, constellation);
                }
                else
                {
                    // not found
                    msg = String.Format("Coordinate RA=[{0}] DEC=[{1}] not found.",
                        txtRA.Text, txtDEC.Text);
                }
                txtResult.Text += msg;
                txtResult.Text += "\r\n";
            }
            else
            {
                // format error
            }
        }

        #endregion

        #region Berechnungen

        /// <summary>
        /// Sucht das Sternbild, in dem die gegebenen Koordinaten liegen.
        /// </summary>
        /// <param name="ra">RA-Koordinate (0°..360°).</param>
        /// <param name="dec">DEC-Koordinate (-90°..+90°).</param>
        /// <param name="name">Name des Sternbild, in dem die Koordinate liegt.</param>
        /// <returns>true, wenn das Sternbild gefunden wurde.</returns>
        private bool FindConstellation(double ra, double dec, ref string name)
        {
            bool found = false;
            Coordinate coordinate = new Coordinate(ra,dec);
            int intercepts = 0;

            if (LoadConstellationBoundaries())
            {
                foreach (string abbreviation in _boundaries.Keys)
                {
                    intercepts = TestCoordinate(coordinate,
                        (Collection<Coordinate>)_boundaries[abbreviation]);
                    if ((intercepts % 2) == 1)
                    {
                        // zur Vereinfachung mal die Pol-Sternbilder ausschließen
                        // (kleiner Wagen im Norden, Oktant in Süden
                        if ((abbreviation != "umi") && (abbreviation != "oct"))
                        {
                            name += abbreviation;
                            name += ";";
                            found = true;
                        }
                    }
                }
            }

            return found;
        }

        /// <summary>
        /// Berechnet aus der gegebenen (zu prüfenden) Koordinate und den
        /// Sternbildgrenzen die Anzahl der Schnittpunkte mit der gedachten Linie.
        /// </summary>
        /// <param name="coordinate">Die zu prüfende Koordinate.</param>
        /// <param name="coll">Eine Liste der Punkte der Sternbildgrenze.</param>
        /// <returns>Die Anzahl der Schnittpunkte der gedachten Gerade mit den Sternbildgrenzen.</returns>
        int TestCoordinate(Coordinate coordinate, Collection<Coordinate> coll)
        {
            int intercepts = 0;
            int points = coll.Count;
            double m1, b1;
            double m2, b2;
            double sx, sy;
            double raMin1, raMax1;
            double decMin1, decMax1;
            double raMin2, raMax2;
            double decMin2, decMax2;

            for (int i = 0; i < points - 1; ++i)
            {
                Coordinate coord1 = coll[i];
                Coordinate coord2 = coll[i + 1];

                // Ist die Länge kleiner als 30 (also 30° Abstand)
                if (Distance(coord1, coord2) < 30)
                {
                    // minima und maxima bestimmen
                    raMin1 = Math.Min(coord1.Rektaszension, coord2.Rektaszension);
                    raMax1 = Math.Max(coord1.Rektaszension, coord2.Rektaszension);
                    decMin1 = Math.Min(coord1.Deklination, coord2.Deklination);
                    decMax1 = Math.Max(coord1.Deklination, coord2.Deklination);
                    raMin2 = Math.Min(coordinate.Rektaszension, _referenzpunkt.Rektaszension);
                    raMax2 = Math.Max(coordinate.Rektaszension, _referenzpunkt.Rektaszension);
                    decMin2 = Math.Min(coordinate.Deklination, _referenzpunkt.Deklination);
                    decMax2 = Math.Max(coordinate.Deklination, _referenzpunkt.Deklination);

                    // erste Gerade m1*x+b1 aus c1 und c2
                    m1 = 10000000000.0;
                    if (coord1.Rektaszension != coord2.Rektaszension)
                    {
                        m1 = (coord2.Deklination - coord1.Deklination)
                           / (coord2.Rektaszension - coord1.Rektaszension);
                    }
                    b1 = coord1.Deklination - coord1.Rektaszension * m1;

                    // zweite Gerade m2*x+b2 aus Such- und Referenzpunkt
                    m2 = 9999999999.0;
                    if (coordinate.Rektaszension != _referenzpunkt.Rektaszension)
                    {
                        m2 = (_referenzpunkt.Deklination - coordinate.Deklination)
                           / (_referenzpunkt.Rektaszension - coordinate.Rektaszension);
                    }
                    b2 = coordinate.Deklination - coordinate.Rektaszension * m2;

                    // Schnittpunkt
                    sx = (b2 - b1) / (m1 - m2);
                    sy = ((m1 * b2) - (m2 * b1)) / (m1 - m2);

                    // ist der Schnittpunkt im Gültigkeitsbereich?
                    if ((sx >= raMin1) && (sx <= raMax1) && (sy >= decMin1) && (sy <= decMax1)
                     && (sx >= raMin2) && (sx <= raMax2) && (sy >= decMin2) && (sy <= decMax2))
                    {
                        ++intercepts;
                    }
                }
            }

            return intercepts;
        }

        /// <summary>
        /// Berechnet den Abstand der beiden Punkte.
        /// </summary>
        /// <param name="c1">Erste Koordinate.</param>
        /// <param name="c2">Zweite Koordinate.</param>
        /// <returns>Der Abstand der beiden Koordinatenpunkte.</returns>
        private double Distance(Coordinate c1, Coordinate c2)
        {
            double distance = Double.NaN;
            double dx = Double.NaN;
            double dy = Double.NaN;

            dx = c1.Rektaszension - c2.Rektaszension;
            dy = c1.Deklination - c2.Deklination;
            distance = Math.Sqrt(dx * dx + dy * dy);

            return distance;
        }

        #endregion

        #region Dateizugriff

        /// <summary>
        /// Lädt die Sternbildgrenzen aus der in den Einstellungen angegebenen
        /// Datei. Diese Datei stammt aus ...
        /// (Beachte das Format)
        /// </summary>
        /// <returns>true, wenn die Daten gelesen wurden.</returns>
        bool LoadConstellationBoundaries()
        {
            bool ok = false;
            string zeile = String.Empty;
            string[] splitted = null;
            string abbreviation = String.Empty;
            int intDummy = Int32.MinValue;
            double ra = Double.NaN;
            double dec = Double.NaN;

            // Daten schon geladen?
            if (!_constellationsLoaded)
            {
                try
                {
                    // Datenstruktur anlegen
                    _boundaries = new Hashtable();

                    // Datei öffnen
                    StreamReader sr = new StreamReader(
                        Properties.Settings.Default.ConstellationBoundariesFile);

                    // über alle Zeilen iterieren
                    while (!sr.EndOfStream)
                    {
                        // Zeile einlesen
                        zeile = sr.ReadLine();

                        // Gültige Datenzeile (erste Prüfung) ?
                        if (zeile.Length == 29)
                        {
                            // Gültige Datenzeile (zweite Prüfung) ?
                            if ((zeile[0] != '#')
                             && (Int32.TryParse(zeile.Substring(0, 3), out intDummy)))
                            {
                                // Daten extrahieren
                                splitted = zeile.Split(' ');
                                ra = Double.Parse(splitted[0].Replace(".", ","));
                                dec = Double.Parse(splitted[1].Replace(".", ","));
                                abbreviation = splitted[2].ToLower();

                                // Daten in Struktur übernehmen
                                Collection<Coordinate> coll =
                                    (Collection<Coordinate>)_boundaries[abbreviation];
                                if (coll == null)
                                {
                                    coll = new Collection<Coordinate>();
                                    _boundaries[abbreviation] = coll;
                                }

                                Coordinate coordinate = new Coordinate(ra, dec);
                                coll.Add(coordinate);
                            }
                        }
                    }
                    sr.Close();
                    _constellationsLoaded = true;
                    ok = true;
                }
                catch (Exception)
                {
                    // Fehler beim Lesen
                    ok = false;
                }
            }
            else
            {
                // Daten sind geladen
                ok = true;
            }

            return ok;
        }

        #endregion
    }
}
